Timers = LibStub("AceAddon-3.0"):NewAddon("Timers","AceConsole-3.0", "AceEvent-3.0");

local MAX_TIMER_ROWS = 10
local TIMERS_CFGVERSION = 3
local NUM_AURAS=41

local DEFAULTS = {
      	       spacing = 4,
	       height = 16,
	       width  = 200,
	       seconds = 15,
	       locked = false,
	       hide = false,
	       combat = false,
	       spec = 'both',
      }

local gHasEnhance = nil;
local gHasCombat = nil;
local gHasSpec = nil;

function Debug(Value)
  DEFAULT_CHAT_FRAME:AddMessage(Value)
end
local timerFrameArgs = {
      name = {
      	   name = 'Name',
	   usage = 'Frame Title',
	   type = 'input',
	   handler = Timers, -- should become self
	   get = "GetValue",
	   set = "SetValue",
	   order = 1,
      },
      height = {
      	     name = 'Height',
	     type= 'range',
	     min = 10,
	     max = 50,
	     step = 1,
	     handler = Timers, -- should become self
	     get = "GetValue",
	     set = "SetValue",
	     },	     
      spacing = {
      	      name = 'Gap',
	      type='range',
	      min=0,	
	      max=30,
	      step=1,
	      handler = Timers, -- should become self
	      get = "GetValue",
	      set = "SetValue",
      },
      width = {
      	      name = 'Width',
	      type='range',
	      min=20,	
	      max=600,
	      step=1,
	      handler = Timers, -- should become self
	      get = "GetValue",
	      set = "SetValue",
      },
      seconds = {
      	      name = 'Seconds',
	      desc = 'Number of seconds for a "full" bar',
	      type='range',
	      min=10,	
--	      max=60*60,
	      max = 300,
	      step=1,
	      handler = Timers, -- should become self
	      get = "GetValue",
	      set = "SetValue",
      },
      locked = {
      	      name = 'Locked',
	      desc = 'Lock frame in place',
	      type='toggle',
	      handler = Timers, -- should become self
	      get = "GetValue",
	      set = "SetValue",
	      order = 3,
      },
      combat = {
            name='Combat Only',
	    desc='Show frame in combat only',
	    type = 'toggle',
	    handler = Timers,
	    get = "GetValue",
	    set = "SetValue",
      },
      spec = {
      	   name = 'Spec',
	   desc = 'Show on which spec',
	   type = 'select',
	   values = {both = 'Both', prim = 'Primary', sec= 'Secondary'},
	    handler = Timers,
	    get = "GetValue",
	    set = "SetValue",
      },
      hide = {
      	      name = 'Hide',
	      desc = 'Hide frame',
	      type='toggle',
	      handler = Timers, -- should become self
	      get = "GetValue",
	      set = "SetValue",
	      order = 4,
      },
      del = {
      	      name = 'Remove',
	      desc = 'Remove frame in place',
	      type='execute',
	      handler = Timers, -- should become self
	      func = "DelFrame",
	      confirm = true,
	      confirmText = "Are you sure you want to remove this group of timers? This can't be undone.",
	      order = 2,
      },
}
local oneTimerArgs = {
      desc = { 
      	   name = function(info) return Timers:GetLabel(info,"Bar %d") end,
	   type = 'description',
	   order = 10,
	   },
	ability = {
	       name = 'Spell/Cooldown',
	       usage = 'Put the spell or ability name here',
	       desc = '',
	       type = 'input',
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 20,
	       },
 	type = {
	       name = 'Shows On',
	       desc = 'Triggered By',
	       type = 'select',
	       values = { off = 'Disabled', dot = 'Debuff on target', cd = 'Cooldown', buff = 'Self', pet = 'Pet', focus = 'Focus', tench = 'Weapon Enchant'},
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 30,
	       },
	mine = {
	       name = 'Mine',
	       desc = 'Only show mine',
	       type = 'toggle',
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 40,
	       },
	count = {
	      name = 'Count',
	      desc = 'Stack Count to turn Active at',
	      type = 'input',
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 41,	      
	},
	invert = {
	       name = 'Invert',
	       desc = 'Invert Colors',
	       type = 'toggle',
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 43,
	       },
	hidei = {
	       name = 'Hide',
	       desc = 'Hide when inactive',
	       type = 'toggle',
	       handler = Timers, -- should become self
	       get = "GetValue",
	       set = "SetValue",
	       order = 47,
	       },
      up = {
      	      name = 'Move Up',
	      desc = 'Move Up One Level',
	      type='execute',
	      handler = Timers, -- should become self
	      func = "MoveUp",
	      order = 50,
      },
      down = {
      	      name = 'Move Down',
	      desc = 'Move Down One Level',
	      type='execute',
	      handler = Timers, -- should become self
	      func = "MoveDown",
	      order = 60,
      },
};


local options = {
      type = 'group',
      args = {
      	bars = {
		type = 'group',
		childGroups = 'tab',
		args = {
		     reset = {
		     	   name = "Reset",
			   desc = "Clear all timers",
			   type = "execute",
			   handler = Timers,
			   func = "Reset",
			   order = 2,
			   },
		     add = {
		     	   name = "Add Frame",
			   desc = "Add a new Frame",
			   type = "execute",
			   handler = Timers,
			   func = "AddFrame",
			   order = 1,
		     },
		},
	    },
	}
}

function Timers:SetOptions()
	-- build options list
	for frame,v in pairs(self.db.profile.Timers.frame) do
	    if (v) then
	       local name = frame
	       if (v[0] and v[0].name) then
	       	  name = v[0].name
	       end
	       options.args.bars.args[frame] = {
			 	type = 'group',
				desc = name,
				name = frame,
--				inline = true,
--				childGroups = "tab",
				args = {},
	       };
	       local i
	       for i=1,MAX_TIMER_ROWS,1 do
	       	   options.args.bars.args[frame].args['entry' .. i] = {
	    						   type = 'group',
							   name = '',
							   args = oneTimerArgs,
							   order = i,
							   inline = true,
	    						   }
	       end
	       options.args.bars.args[frame].args.cfg = {
					       type = 'group',
					       name = '',
					       inline = true,
					       order = 0,
					       args = timerFrameArgs,
	    				       }
	   end
	end
	self.options_debug = options
end

function Timers:OnInitialize()
	self.defaults = {
	      profile = {
	      	      Timers = { frame = nil },
	      }
	}
	self.db = LibStub("AceDB-3.0"):New("TimersData", self.defaults,"class");
	if ((self.db.profile.cfgversion == nil) or (self.db.profile.cfgversion < TIMERS_CFGVERSION)) then
	   self.db:ResetProfile()
	   self.db.profile.cfgversion = TIMERS_CFGVERSION
	   self:Print("Major configuration updates - configuration reset (SORRY)");
	end
--	self:SetOptions()
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Timers", options.args.bars)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Timers Profile",options.args.profile)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Timers","Timers")
	self.profileFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Timers Profile","Profile","Timers")
	self.db.RegisterCallback(self, 'OnNewProfile', 'OnProfileReset')
	self.db.RegisterCallback(self, 'OnProfileChanged', 'OnProfileReset')
	self.db.RegisterCallback(self, 'OnProfileCopied', 'OnProfileReset')
	self.db.RegisterCallback(self, 'OnProfileReset', 'OnProfileReset')
	self.db.RegisterCallback(self, 'OnProfileDeleted', 'OnProfileReset')

	self:RegisterChatCommand("timers","ChatCommand")
	local ver = GetAddOnMetadata("Timers","version");
	if (ver) then
	   self:Print("Timers v" .. ver .. " Initialized")
	else
	   self:Print("Timers Initialized")
	end
end 

function Timers:OnProfileReset()
	 self:Reset()
	 for k,v in pairs(self.data.frames) do
	     v:Hide()
	 end
	 self:LoadProfile()
end

function Timers:ApplySettings()
	 self:LoadProfile()
end


function Timers:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub("AceConfigCmd-3.0").HandleCommand(Timers,"timers","Timers",input)
    end
end

function Timers:GetActualValue(frame,level,name)
	 if (not self.db.profile.Timers.frame) then self.db.profile.Timers.frame = {}; end
	 if (not self.db.profile.Timers.frame[frame]) then self.db.profile.Timers.frame[frame] = {}; end
	 if (not self.db.profile.Timers.frame[frame][level]) then self.db.profile.Timers.frame[frame][level] = {}; end
	 return self.db.profile.Timers.frame[frame][level][name]
end

function Timers:SetActualValue(frame,level,name,value)
	 if (not self.db.profile.Timers.frame) then self.db.profile.Timers.frame = {}; end
	 if (not self.db.profile.Timers.frame[frame]) then self.db.profile.Timers.frame[frame] = {}; end
	 if (not self.db.profile.Timers.frame[frame][level]) then self.db.profile.Timers.frame[frame][level] = {}; end
	 self.db.profile.Timers.frame[frame][level][name] = value
end

function Timers:GetLastLevel(info,level)
	 if (level<=1) then return info.options; end
	 if (not info) then return nil; end
	 local options = self:GetLastLevel(info,level-1)
	 if (options and options.args) then return options.args[info[level-1]] end
	 return nil
end

function Timers:IsSpec(spec)
	 if (spec == 'both') then return true; end
	 local actSpec = GetActiveTalentGroup(false,false);
	 if (spec == 'prim' and actSpec == 1) then return true; end;
	 if (spec == 'sec' and actSpec == 2) then return true; end;
	 return false;
end

function Timers:IsShown(frame)
	 local config =  self.db.profile.Timers.frame[frame][0];
	 if (config.hide) then return false; end
	 if (config.combat and not UnitAffectingCombat("player")) then 
	    return false;
	 end
	 if (not self:IsSpec(config.spec)) then 
	    return false;
	 end
	 return true;
end

function Timers:GetValue(info)
	 local frame = self:GetLastLevel(info, #info-1).name
	 local name = info[#info]
	 local level = Timers:GetLastLevel(info,#info)
	 return self:GetActualValue(frame,level.order,name)
end
function Timers:SetValue(info,value)
	 local frame = self:GetLastLevel(info, #info-1).name
	 local name = info[#info]
	 local level = Timers:GetLastLevel(info,#info)
	 self:SetActualValue(frame,level.order,name,value)
	 -- Special case this!
--	 if (level == 0 and name == "name") then   -- changing frame name
--	    ;
--	 end
--	 Special cases which don't need full updates
	 if (name == "combat") then
	        if (value) then 
		   gHasCombat = true; 
		else
		   gHasCombat = false;
		   for frame,v in pairs(self.db.profile.Timers.frame) do
	     		if (self.db.profile.Timers.frame[frame][0].combat) then
	     		   gHasCombat=true;
	     		end
	 	   end
		end
	 end
	 if (name == "spec") then
	        if (value) then 
		   gHasSpec = true; 
		else
		   gHasSpec = false;
		   for frame,v in pairs(self.db.profile.Timers.frame) do
	     		if (self.db.profile.Timers.frame[frame][0].spec ~= 'both') then
	     		   gHasSpec=true;
	     		end
	 	   end
		end
	 end
	 if (name=="combat" or name == "spec" or name == "hide") then
	         self:ShowOrHide(frame)	    		   
	 else
	 	 self:Update()
         end
end

function Timers:GetLabel(info,value)
	 local level = Timers:GetLastLevel(info,#info)
	 if (level) then return(string.format(value,level.order)) else return value end;
end

function Timers:Update()
	 self:Reset()
	 self:LoadProfile()
end

function Timers:Reset()
	 if (not self.data) then return; end
	 for k,v in pairs(self.data.abilities) do
	     if (v) then 
	     	for i,j in pairs(v) do
		    if (j) then 
		       for l,entry in pairs(j) do
		       	   if (entry) then
			      entry.bar:Hide()
			      entry.bar.frame.timeText = nil
			      entry.bar.frame.text = nil
			      entry.bar.frame = nil
			      entry.bar.icon = nil
			   end
--		       	   if (entry) then CB:Unregister(entry.timer); end
			   -- should do something here... 
			   -- lots of lost data/frames
			   -- garbage collection or something
		       end
		    end
		end
	     end
	 end
	 self.data = {frames = self.data.frames, abilities = {}}
end

function Timers:AddFrame()
	 local numsUsed = {};
	 local found = 0;
	 for k,v in pairs(self.db.profile.Timers.frame) do
	     local num = string.match(k,"^Frame(%d+)$");
	     numsUsed[tonumber(num)] = true;
	     found = found + 1;
	 end
	 local num = found + 1;
	 local i;
	 for i=1, found, 1 do
	     if (not numsUsed[i]) then num = i; end
	 end
	 self.db.profile.Timers.frame["Frame" .. num] = {}
	 self:Update()
	 self:SetOptions()
end
function Timers:DelFrame(info)
	 local frame = self:GetLastLevel(info, #info-1).name
	 self.db.profile.Timers.frame[frame] = nil;
	 options.args.bars.args[frame] = nil;
	 if (self.data.frames[frame]) then self.data.frames[frame]:Hide(); end
	 self:Update()
	 self:SetOptions()
end

function Timers:Swap(frame,level1,level2)
	 local tmp = self.db.profile.Timers.frame[frame][level1]
	 self.db.profile.Timers.frame[frame][level1] = self.db.profile.Timers.frame[frame][level2]
	 self.db.profile.Timers.frame[frame][level2] = tmp
end

function Timers:MoveUp(info)
	 local frame = self:GetLastLevel(info, #info-1).name
	 local level = Timers:GetLastLevel(info,#info).order
	 if (level==1) then return; end
	 self:Swap(frame,level,level-1)
	 self:Update()
	 self:SetOptions()
end
function Timers:MoveDown(info)
	 local frame = self:GetLastLevel(info, #info-1).name
	 local level = Timers:GetLastLevel(info,#info).order
	 if (level==MAX_TIMER_ROWS) then return; end
	 self:Swap(frame,level,level+1)
	 self:Update()
	 self:SetOptions()
end

function Timers:Upperfirst(s)
	if (s == nil) then return s; end
	local f = string.upper(string.sub(s,1,1))
	return f .. string.sub(s,2)
end

function Timers:Create(origEntry,position,spacing,height,width,seconds,frame,frameName)
	 local entry = {}
	 local k,v
	 for k,v in pairs(origEntry) do
	     entry[k] = v
	 end
	 entry.seconds = seconds
	 if (not entry.timerText or entry.timerText == "") then entry.timerText = entry.ability; end
	 if (not entry.timer or entry.timer == "") then entry.timer = entry.ability .. entry.type .. frameName .. position; end
	 entry.bar = CreateFrame("Frame",nil,frame)
	 entry.bar:SetHeight(height)
	 entry.bar:SetWidth(width+height)
	 entry.bar:SetPoint("TOPLEFT",0,-1 - (spacing+height)*position)
	 entry.bar.icon = CreateFrame("Button",nil,entry.bar)
	 entry.bar.icon:SetHeight(height)
	 entry.bar.icon:SetWidth(height)
	 entry.bar.icon:SetPoint("LEFT")
	 entry.bar.frame = CreateFrame("StatusBar",nil,entry.bar)
	 entry.bar.frame:SetHeight(height)
	 entry.bar.frame:SetWidth(width)
	 entry.bar.frame:SetPoint("RIGHT")
	 entry.bar.frame:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar");
	 entry.bar.frame:SetMinMaxValues(0,seconds)
	 entry.bar.frame:SetValue(seconds)
-- need this in config variables
	 local font, _, style = GameFontHighlight:GetFont()
	 local fontsize = 11;
	 entry.bar.frame.timeText = entry.bar.frame:CreateFontString(nil,"OVERLAY")	 
	 entry.bar.frame.timeText:SetFontObject(GameFontHighlight)
	 entry.bar.frame.timeText:SetFont(font,fontsize,style)
	 entry.bar.frame.timeText:SetHeight(height)
	 entry.bar.frame.timeText:SetWidth(fontsize*3.6)
	 entry.bar.frame.timeText:SetPoint("LEFT",entry.bar.frame,"LEFT")
	 entry.bar.frame.timeText:SetJustifyH("RIGHT")
	 entry.bar.frame.timeText:SetText("0.0")
	 entry.bar.frame.timeText:SetTextColor(1,1,1,1)
	 entry.bar.frame.text = entry.bar.frame:CreateFontString(nil,"OVERLAY")	 
	 entry.bar.frame.text:SetFontObject(GameFontHighlight)
	 entry.bar.frame.text:SetFont(font, fontsize, style)
	 entry.bar.frame.text:SetHeight(height)
	 entry.bar.frame.text:SetWidth(width-height-(fontsize*3.6))
	 entry.bar.frame.text:SetPoint("LEFT",entry.bar.frame.timeText,"RIGHT",4,0)
	 entry.bar.frame.text:SetJustifyH("LEFT")
	 entry.bar.frame.text:SetText(entry.timerText)
	 entry.bar.frame.text:SetTextColor(1,1,1,1)

	 if (entry.invert) then
	        entry.bar.frame:SetStatusBarColor(0,1,0,1)
	 else
	        entry.bar.frame:SetStatusBarColor(1,0,0,1)
    	 end
	 if (entry.hidei) then
	    entry.bar:Hide()
	 else
	    entry.bar:Show()
--	    CB:SetTimeLeft(entry.timer,1)
--	    CB:Start(entry.timer)
--	    CB:SetTimeLeft(entry.timer,0)
--	    CB:Update(entry.timer)
	 end
	 self:SetIcon(entry,GetSpellTexture(entry.ability))
	 if (not self.data) then self.data = {frames = {}, abilities = {}}; end
	 if (not self.data.abilities[entry.type]) then self.data.abilities[entry.type] = {}; end
	 if (not self.data.abilities[entry.type][entry.ability]) then self.data.abilities[entry.type][entry.ability] = {}; end
	 table.insert(self.data.abilities[entry.type][entry.ability],entry)
end

function Timers:ShowFrame(frame)
	 self.data.frames[frame]:Show()
end
function Timers:HideFrame(frame)
    	 self.data.frames[frame]:Hide()
end

function Timers:LockFrame(frame)
	 self.data.frames[frame].fontFrame:SetBackdrop(nil)
--    	 self.data.frames[frame].fontFrame:EnableMouse(false)
    	 self.data.frames[frame].fontFrame:EnableMouse(true)
end

function Timers:UnlockFrame(frame)
	 self.data.frames[frame].fontFrame:SetBackdrop({
	   bgFile="Interface\\Tooltips\\UI-Tooltip-Background", 
	   edgeFile="Interface\\Tooltips\\UI-Tooltip-Border", 
	   tile=1, tileSize=10, edgeSize=10, 
	   insets={left=0, right=0, top=0, bottom=0}
	 });
    	 self.data.frames[frame].fontFrame:EnableMouse(true)
end

function Timers:SavePos(frame)
	 point, relativeTo, relativePoint, offsetX, offsetY = self.data.frames[frame]:GetPoint(1)
	 self.db.profile.Timers.frame[frame][0].rect = { point,relativeTo,relativePoint,offsetX,offsetY };
end

function Timers:StartMoving(actFrame,locked)
	  if (actFrame) then
	     	  if ((not locked) or IsAltKeyDown()) then
		    	  actFrame:StartMoving()
                  end
	  end
end

function Timers:LoadFrame(frame)
	 if (not self.db.profile.Timers.frame or not self.db.profile.Timers.frame[frame]) then return; end
	 if (not self.data) then self.data = {frames = {}, abilities = {}}; end
	 if (not self.db.profile.Timers.frame[frame][0]) then self.db.profile.Timers.frame[frame][0] ={}; end
	 -- Just realized the code below is inacurate for boolean values
	 -- right now all bools default to false so it works, but needs to be fixed
	 local config =  self.db.profile.Timers.frame[frame][0];
	 local spacing = config.spacing or DEFAULTS.spacing
	 local height = config.height or DEFAULTS.height
	 local seconds = config.seconds or DEFAULTS.seconds
	 local width = config.width or DEFAULTS.width
	 local locked = config.locked or DEFAULTS.locked
	 local hide = config.hide or DEFAULTS.hide
	 local name = config.name or frame
	 local combat = config.combat or DEFAULTS.combat;
	 local spec = config.spec or DEFAULTS.spec;
	 config.spacing=spacing
	 config.height=height
	 config.seconds=seconds
	 config.width=width
	 config.locked=locked
	 config.name=name
	 config.hide=hide
	 config.combat=combat
	 config.spec = spec
	 if (combat) then gHasCombat = true; end;
	 if (spec ~= 'both') then gHasSpec = true; end;
	 if (not self.data.frames[frame]) then
	    local actFrame = CreateFrame("Frame","TimersFrame" .. frame,UIParent)
	    actFrame:SetFrameStrata("BACKGROUND")
	    actFrame:SetWidth(1) -- tbd
	    actFrame:SetHeight(1) --tbd
	    if (config.rect) then
	        actFrame:SetPoint(config.rect[1],config.rect[2],config.rect[3],config.rect[4],config.rect[5])
	    else
		actFrame:SetPoint("CENTER",0,0) -- get this from prifle
	    end
	    -- create title frame here
	    actFrame.fontFrame = CreateFrame("Frame","$parentTitle",actFrame)
	    actFrame.fontFrame:SetWidth(width+height)
	    actFrame.fontFrame:SetHeight(15)
	    actFrame.fontFrame:SetPoint("TOPLEFT",actFrame,"TOPLEFT",0,0); -- offset by height for icon

	    actFrame.fontFrame.title = actFrame.fontFrame:CreateFontString("$parentText","ARTWORK","GameFontNormal");
	    actFrame.fontFrame.title:SetAllPoints();
	    actFrame.fontFrame.title:SetText(name)
	    actFrame:SetMovable(true)
	    actFrame.fontFrame:EnableMouse(false)
	    actFrame.fontFrame:SetScript("OnMouseDown",function() self:StartMoving(actFrame,locked); end)
	    actFrame.fontFrame:SetScript("OnMouseUp",function() actFrame:StopMovingOrSizing(); self:SavePos(frame) end)
	    self.data.frames[frame] = actFrame
	 else
	    self.data.frames[frame].fontFrame:SetWidth(width+height)
	    self.data.frames[frame].fontFrame.title:SetText(name)
	    if (config.rect) then
	        self.data.frames[frame]:SetPoint(config.rect[1],config.rect[2],config.rect[3],config.rect[4],config.rect[5])
	    end
	 end
	 if (locked) then 
	    self:LockFrame(frame)
	 else
	    self:UnlockFrame(frame)
	 end
	 for pos, entry in pairs(self.db.profile.Timers.frame[frame]) do
	     if (entry and entry.type and entry.type ~= 'off' and entry.ability and entry.ability ~= '') then
	     	     self:Create(entry,pos,spacing,height,width,seconds,self.data.frames[frame],frame)
		     if (entry.type == 'tench') then gHasEnhance = true; end;
             end
	 end
	 local inCombat = UnitAffectingCombat("player")
	 self:ShowOrHide(frame)
end

function Timers:ShowOrHide(frame)
	 if (self:IsShown(frame)) then
	        self:ShowFrame(frame)
	 else
		self:HideFrame(frame)
	 end
end

function Timers:LoadProfile()
	 gHasEnhance = nil;
	 gHasCombat = nil;
	 gHasSpec = nil;
	 if (not self.db.profile.Timers.frame) then 
	    self.db.profile.Timers.frame = { Frame1 = {} }
	 end
	 for frame,v in pairs(self.db.profile.Timers.frame) do
	     self:LoadFrame(frame)
	 end
--	 self:checkDebuffs()
	 self:checkCooldowns()
	 self:checkDebuffs("target",self.data.abilities.dot)
	 local seen = {}
	 self:checkBuffs("player",self.data.abilities.buff,seen)
	 self:checkDebuffs("player",self.data.abilities.buff,seen)
	 self:checkBuffs("pet",self.data.abilities.pet)
	 seen = {}
	 self:checkBuffs("focus",self.data.abilities.focus,seen) 
	 self:checkDebuffs("focus",self.data.abilities.focus,seen) 
	 self:checkWeapons()
end

function Timers:OnEnable()
	 if (not self.db.profile.Timers.frame) then 
	    self.db.profile.Timers.frame = { Frame1 = {} }
	 end
	 if (self.db.profile.Timers.frame.Timers and not self.db.profile.Timers.frame.Frame1) then
	    local tmp = self.db.profile.Timers.frame.Timers
	    self.db.profile.Timers.frame = { Frame1 = tmp }
	 end
	 self:SetOptions()
	 self:LoadProfile()

       	 self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	 self:RegisterEvent("SPELLS_CHANGED","SPELL_UPDATE_COOLDOWN")
	 self:RegisterEvent("PLAYER_TARGET_CHANGED")
	 self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	 self:RegisterEvent("UNIT_AURA")
	 self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	 self:RegisterEvent("PLAYER_REGEN_ENABLED")
	 self:RegisterEvent("PLAYER_REGEN_DISABLED")
	 self:RegisterEvent("PLAYER_TALENT_UPDATE")
	 self:RegisterEvent("BAG_UPDATE")
	 if (not self.updFrame) then
	     self.updFrame = CreateFrame("Frame")
	     self.updFrame:Show()
    	     self.updFrame:SetScript("OnUpdate",function() self:OnUpdate(); end)	 end

end
function Timers:OnDisable()
       	 self:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
	 self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	 self:UnregisterEvent("PLAYER_FOCUS_CHANGED")
	 self:UnregisterEvent("SPELLS_CHANGED")
	 self:UnregisterEvent("UNIT_AURA")
	 self:Reset()
	 for k,v in pairs(self.data.frames) do
	     v:Hide()
	 end
end
function Timers:SetIcon(entry,icon)
	 if (not entry or not icon) then return; end
	 entry.bar.icon:SetNormalTexture(icon)
         entry.bar.icon:GetNormalTexture():SetTexCoord(.07,.93,.07,.93)
end

function Timers:checkAuras(buff,target,auraList,seen)
	 if (not self.data) then return; end
	 if (not auraList) then return; end
	 if (not seen) then seen = {}; end
	 local func = UnitDebuff;
	 if (buff) then 
	    func=UnitAura;
         end
	 for i=1,NUM_AURAS,1 do
	 	 local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = func(target,i)
		 if (isMine ~= "player" and isMine ~= "vehicle") then isMine = nil; end;
		 local entryList = auraList[name]
		 if (entryList) then
		    if (isMine) then	
		       seen[name] = 2
		    elseif (not seen[name]) then
		       seen[name] = 1
		    end		    
		    for k,entry in pairs(entryList) do 
		      if (entry and ((not entry.mine) or isMine)) then
		     	self:StartTimer(entry,expirationTime,count)
		        self:SetIcon(entry,icon);
		      end
                    end
		 end
	 end
	 -- now iterate through existing items
	 
	 for k,entryList in pairs(auraList) do
	     for l,entry in pairs(entryList) do 
	     	 if (not seen[k] or (entry.mine and seen[k]==1)) then
		    self:StopTimer(entry)
		 end
	     end
	 end
end

function Timers:checkDebuffs(target,auraList,seen)
	 self:checkAuras(false,target,auraList,seen)
end
function Timers:checkBuffs(target,auraList,seen)
	 self:checkAuras(true,target,auraList,seen)
end

function Timers:checkCooldowns()
	 if (not self.data or not self.data.abilities or not self.data.abilities.cd) then return; end
	 -- changing this logic now to search for items we have
	 local funcs = {
	       spell = {  usable=IsUsableSpell, cooldown = GetSpellCooldown,count=nil },
	       item  = {  usable=IsUsableItem, cooldown = GetItemCooldown, count=GetItemCount },
	 }
	 for spell,entryList in pairs(self.data.abilities.cd) do
	    local cdType='spell'
--	    local __, __, icon, __ = GetSpellInfo(entry.ability)
	    local __, __, icon, __ = GetSpellInfo(spell)
	    if (not icon) then -- assume its an item
--	       icon = GetItemIcon(entry.ability)
	       icon = GetItemIcon(spell)
	       cdType = 'item'
	    end
	    if (icon) then
	       for k,entry in pairs(entryList) do
	       	   if (cdType == 'item') then
			entry.isItem = true;	       
		   else
			entry.isItem = false;
		   end
	       	   self:SetIcon(entry,icon)
	       	   local usable, nomana = funcs[cdType].usable(entry.ability);
	       	   local start,dur,enabled = funcs[cdType].cooldown(entry.ability)
		   local count=nil;
		   if (funcs[cdType].count) then count=funcs[cdType].count(entry.ability,false,true); end;
       	           if (dur > 1.5 and enabled and GetTime() - start < dur ) then
	       	      	 self:StartTimer(entry,dur + start,count)
			 if (not usable and not nomana) then
	       	      	    self:Inactive(entry,count)
			 end
	       	   else 
			 if (not usable and not nomana) then
	       	      	    self:Inactive(entry,count)
			 else
			    self:StopTimer(entry)
			    self:Active(entry,count)
			 end
	           end
		end
            else
	       -- most likely an item we don't have yet
	       -- or a pet that doesn't exist
	       for k,entry in pairs(entryList) do
	          self:StopTimer(entry)
	          self:Inactive(entry,0)
	       end
	    end
	 end
end

function Timers:PLAYER_TALENT_UPDATE()
	 if (not gHasSpec) then return; end;
	 for frame,v in pairs(self.db.profile.Timers.frame) do
	     if (self.db.profile.Timers.frame[frame][0].spec ~= 'both') then
	     	self:ShowOrHide(frame)
	     end
	 end
end

function Timers:PLAYER_REGEN_DISABLED()
	 if (not gHasCombat) then return; end;
	 for frame,v in pairs(self.db.profile.Timers.frame) do
	     if (self.db.profile.Timers.frame[frame][0].combat) then
	     	self:ShowOrHide(frame)
	     end
	 end
end

function Timers:PLAYER_REGEN_ENABLED()
	 if (not gHasCombat) then return; end;
	 for frame,v in pairs(self.db.profile.Timers.frame) do
	     if (self.db.profile.Timers.frame[frame][0].combat) then
	     	self:ShowOrHide(frame)
	     end
	 end
end

function Timers:getSlotEnchant(slotId)
	 if (not self.data.tip) then 
	    self.data.tip = CreateFrame("GameTooltip","TimersScanningTooltip",nil,"GameTooltipTemplate");
	    self.data.tip:SetOwner( WorldFrame, "ANCHOR_NONE" );
--	    self.data.tip:SetOwner( UIParent, "ANCHOR_NONE" );
	 end
	 self.data.tip:ClearLines();
	 self.data.tip:SetInventoryItem("player",slotId);
	 local line = nil;
	 for i=6, self.data.tip:NumLines() do
	     line = _G["TimersScanningTooltipTextLeft" .. i ]:GetText():match("^(.+) %(%d+ .+%)$");
	     if (line) then 
	     	return line
	     end
	 end
end

function Timers:checkWeapons()
	 if (not (self.data and self.data.abilities and self.data.abilities and self.data.abilities.tench) ) then return end;
	 local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	 mh = self.data.abilities.tench.main
	 local mainHandName = "No Mainhand"
	 if (hasMainHandEnchant) then
	    mainHandName = self:getSlotEnchant(16) or "mainhand"
	 end
	 local offHandName = "No Offhand"
	 if (hasOffHandEnchant) then
	    offHandName = self:getSlotEnchant(17) or "offhand"
	 end
	 if (mh) then
	    for k,entry in pairs(mh) do
	    	entry.timerText = mainHandName;
	    	if (hasMainHandEnchant) then
	       	   self:StartTimer(entry,GetTime() + mainHandExpiration / 1000)
	    	else
		   self:StopTimer(entry)
		end
		self:SetText(entry)
	    end
	 end
	 oh = self.data.abilities.tench.off
	 if (oh) then
	    for k,entry in pairs(oh) do
	    	entry.timerText = offHandName;
	    	if (hasOffHandEnchant) then
	       	   self:StartTimer(entry,GetTime() + offHandExpiration / 1000)
	    	else
		   self:StopTimer(entry)
		end
		self:SetText(entry)
	    end
	 end
end

function Timers:OnUpdate()
	 if (not self.data or not self.data.abilities) then return; end
	 if (self.data.lastCheck and self.data.lastCheck > GetTime() - .1) then return; end
	 for k,v in pairs(self.data.abilities) do
	     if (v) then 
	     	for i,j in pairs(v) do
		    if (j) then 
		       for l,entry in pairs(j) do
		       	   if (entry and entry.bar and entry.bar.expire) then 
			      self:UpdateTimer(entry)
			   end
		       end
		    end
		end
	     end
	 end
	 if (self.data.checkWeapons and self.data.checkWeapons < GetTime()) then
	    self:checkWeapons()
	    self.data.checkWeapons = nil;
	 end
	 self.data.lastCheck = GetTime()
end

function Timers:SPELL_UPDATE_COOLDOWN(event,arg1,arg2,arg3)
	 self:checkCooldowns()
end

function Timers:BAG_UPDATE(event)
	 self:checkCooldowns();
end

function Timers:UNIT_INVENTORY_CHANGED(event,arg1)
	 if (arg1 ~= "player") then return; end
	 if (not gHasEnhance) then return; end
	 self.data.checkWeapons = GetTime() + 2
--	 self:checkWeapons()
end
function Timers:PLAYER_TARGET_CHANGED(event)
	 if (not self.data) then return; end
 	 self:checkDebuffs("target",self.data.abilities.dot)
	 self:checkCooldowns()
end
function Timers:PLAYER_FOCUS_CHANGED(event)
	 if (not self.data) then return; end
	 local seen = {}
 	 self:checkBuffs("focus",self.data.abilities.focus,seen)  
 	 self:checkDebuffs("focus",self.data.abilities.focus,seen) 
end
function Timers:UNIT_AURA(event,unitID)
	 local seen = {}
	 if (not self.data) then return; end
	 if (unitID == "target") then self:checkDebuffs("target",self.data.abilities.dot) end
	 if (unitID == "player") then 
	    self:checkBuffs("player",self.data.abilities.buff,seen) 
	    self:checkDebuffs("player",self.data.abilities.buff,seen) 
	  end
	 if (unitID == "pet") then self:checkBuffs("pet",self.data.abilities.pet) end
-- there's a bug right now where if focus=target, player or pet these aren't checked
--	 if (unitID == "focus") then 
	    self:checkBuffs("focus",self.data.abilities.focus,seen)  
	    self:checkDebuffs("focus",self.data.abilities.focus,seen) 
--	 end
end
function Timers:StartTimer(entry,expire,count)
	 local div = 1
	 if (not count) then count = 0; end
	 local actcount = tonumber(entry.count)
	 if (not actcount or type(actcount)~="number") then actcount = 0; end
	 if (actcount > 0) then
	    div = count / actcount
	    if (div > 1) then div = 1; end
	 end
	 if (entry.invert) then
		entry.bar.frame:SetStatusBarColor(1*div,1*(1-div),0,1)
	 else
		entry.bar.frame:SetStatusBarColor(1*(1-div),1*div,0,1)
	 end
	 entry.bar.expire = expire;
	 entry.bar:Show()
	 self:SetText(entry,count)
	 self:UpdateTimer(entry)
end

function Timers:FormatTime(left)
	 if (not left or not type(left)=='number' or left < 0) then left=0; end
	 if (left < 10) then 
	    return(string.format("%01.1f",left))
	 elseif (left < 60) then 
	    return(string.format("%d",left))
	 else
	    return(string.format("%d:%02d",left / 60, left % 60))
	 end
end

function Timers:UpdateTimer(entry)
	 if (not entry) then return; end
	 if (not entry.bar.expire) then return; end
	 local left = entry.bar.expire - GetTime()
	 self:SetTimeText(entry,left)
	 if (left < 0) then
	    self:StopTimer(entry)
	 else
	    entry.bar.frame:SetValue(left)
	 end
end

function Timers:ShouldShowCount(entry,count)
	 if (not entry) then return false; end
	 count=tonumber(count)
	 local ecount = tonumber(entry.count)	 
	 if (count and count > 1) then 
	    return true
	 end
	 if (entry.type=='cd' and ecount and ecount > 0) then
	    return true;
	 end
	 if (ecount and ecount > 0 and count and count > 0) then
	    return true;
	 end
	 return false;
end

function Timers:SetText(entry,count)
	 if (self:ShouldShowCount(entry,count)) then
	   if (not count) then count=0; end
	   entry.bar.frame.text:SetText(entry.timerText .. " (" .. count .. ")")
 	 else
	   entry.bar.frame.text:SetText(entry.timerText)
	 end
end

function Timers:SetTimeText(entry,left)
	 if (not entry) then return; end
	 entry.bar.frame.timeText:SetText(self:FormatTime(left))
end

function Timers:StopTimer(entry)
	 entry.bar.expire = nil
	 if (entry.hidei) then
	    entry.bar:Hide()
	    return
	 end
	 entry.bar.frame:SetValue(entry.seconds)
	 self:SetTimeText(entry,0)
	 self:SetText(entry)
	 if (entry.invert) then
	   entry.bar.frame:SetStatusBarColor(0,1,0,1)
	 else
	   entry.bar.frame:SetStatusBarColor(1,0,0,1)
	 end
end

function Timers:Inactive(entry,count)
	 self:SetText(entry,count)
	 if (entry.invert) then
	   entry.bar.frame:SetStatusBarColor(0,.5,0,.5)
	 else
	   entry.bar.frame:SetStatusBarColor(.5,0,0,.5)
	 end
end
function Timers:Active(entry,count) 
	 self:SetText(entry,count)
	 if (entry.invert) then
	   entry.bar.frame:SetStatusBarColor(0,1,0,1)
	 else
	   entry.bar.frame:SetStatusBarColor(1,0,0,1)
	 end
end

